from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.atom import *
from meteor_reasoner.classes.rule import *
import copy


def strengthening_transformation(literal):
    new_operators = []
    if isinstance(literal, BinaryLiteral):
        operator = literal.operator
        if operator.name == "Since":
            operator.name = "Diamondminus"
        else:
            operator.name = "Diamondplus"

        if isinstance(literal.right_atom, Atom):
             literal = Literal(literal.right_atom, [operator])
        else:
            literal = literal.right_atom
            literal.operators.insert(0, operator)

    if isinstance(literal, Atom):
        return literal

    for operator in literal.operators:
        if operator.name == "Boxplus":
            operator.name = "Diamondplus"
            new_operators.append(operator)
        elif operator.name == "Boxminus":
            operator.name = "Diamondminus"
            new_operators.append(operator)
        else:
            new_operators.append(operator)

    literal.operators = new_operators[:]
    return literal


def transformation(rules):
    new_rules = []

    for rule in rules:
        head = rule.head
        new_body = []
        for literal in rule.body:
            literal = strengthening_transformation(literal)
            new_body.append(literal)
        rule = Rule(head, new_body)
        new_rules.append(rule)
    return new_rules


def construct_pair(rules):
    pairs = set()
    for rule in rules:
        head = rule.head
        if isinstance(head, Atom):
            origin = [0, 0]
        else:
            origin = [0, 0]
            for operator in head.operators:
                if operator.name == "Boxplus":
                    origin[0] = origin[0] + operator.interval.left_value
                    origin[1] = origin[1] + operator.interval.right_value
                else:
                    origin[0] = origin[0] - operator.interval.left_value
                    origin[1] = origin[1] - operator.interval.right_value

        for literal in rule.body:
            tmp_origin = copy.deepcopy(origin)
            if not isinstance(literal, Atom):
                for operator in literal.operators:
                    if operator.name == "Diamondminus":
                        tmp_origin[0] = tmp_origin[0] + operator.interval.left_value
                        tmp_origin[1] = tmp_origin[1] + operator.interval.right_value
                    if operator.name == "Diamondplus":
                        tmp_origin[0] = tmp_origin[0] - operator.interval.left_value
                        tmp_origin[1] = tmp_origin[1] - operator.interval.right_value
            else:
                tmp_origin[0] = tmp_origin[0] + 0
                tmp_origin[1] = tmp_origin[1] + 0


            pairs.add((literal.get_predicate(), head.get_predicate(), tmp_origin[0], tmp_origin[1]))

    return pairs








from collections import defaultdict
from meteor_reasoner.classes.term import *
from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.atom import *
from meteor_reasoner.materialization.ground import *


def test_ground():
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["C"][tuple([Term("mike"), Term("jack")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]

    literal_a = Literal(Atom("A", tuple([Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False))])

    for entity, intervals, delta in ground_unary_literal(literal=literal_a, context=dict(), D=D):
        assert ["mike"] == [str(item) for item in entity]
        assert ['[3,4]', '(6,10)'] == [str(item) for item in intervals]

    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    for entity, intervals, delta in ground_unary_literal(literal=literal_b, context=dict(), D=D):
        assert ["nan"] == [str(item) for item in entity]
        assert ['[2,8]'] == [str(item) for item in intervals]

    literal_c = Literal(Atom("A", tuple([Term("mike")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    for entity, intervals, delta in ground_unary_literal(literal=literal_c, context=dict(), D=D):
        assert ["mike"] == [str(item) for item in entity]
        assert ['[3,4]', '(6,10)'] == [str(item) for item in intervals]

    literal_d = Literal(Atom("C", tuple([Term("mike"), Term("X", "variable")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    for entity, intervals, delta in ground_unary_literal(literal=literal_d, context=dict(), D=D):
        assert ['mike', 'jack'] == [str(item) for item in entity]
        assert ['[3,4]', '(6,10)'] == [str(item) for item in intervals]




from meteor_reasoner.classes.literal import *
from functools import reduce
from meteor_reasoner.classes.atom import *
from meteor_reasoner.classes.interval import *
import math
from decimal import Decimal


def construct_left_right_pattern(points, gcd):
    """
    This function build two dictionaries used for finding the next interval when
    executing the left or right moving operation.
    Args:
        points (list of distinct points in the dataset):
        gcd (the great common divisor in the program):

    Returns:
        two dictionary objects
    """
    points = list(points)
    points.sort()

    left_end_point = points[0]
    right_end_point = points[0] + gcd

    pattern = set()
    pattern.add(left_end_point)
    pattern.add(right_end_point)
    # the pattern's range [x, x+gcd], all other points will fall into this range
    for item in points:
        a, b = divmod(item-left_end_point, gcd)
        if a == 0:
            continue
        else:
            pattern.add(b)

    pattern = list(pattern)
    pattern.sort()

    intervals = []
    for i in range(len(pattern) - 1):
        item = pattern[i + 1] - pattern[i]
        intervals.append(item)

    right_interval_search_dict = dict()
    pattern = []
    for i in range(len(intervals)):
        pattern.append(str(intervals[i]))
        if i == len(intervals) - 1:
            right_interval_search_dict["#".join(pattern)] = intervals[0]
        else:
            right_interval_search_dict["#".join(pattern)] = intervals[i + 1]

    left_interval_search_dict = dict()
    pattern = []
    for i in range(len(intervals) - 1, -1, -1):
        pattern = [str(intervals[i])] + pattern
        if i == 0:
            left_interval_search_dict["#".join(pattern)] = intervals[len(intervals) - 1]
        else:
            left_interval_search_dict["#".join(pattern)] = intervals[i - 1]

    if len(left_interval_search_dict) == 0:
        left_interval_search_dict["1.0"] = 1
    if len(right_interval_search_dict) == 0:
        right_interval_search_dict["1.0"] = 1
    return left_interval_search_dict, right_interval_search_dict



def interval_intesection_intervallist(target, interval_list):
    """
    Check whether an Interval instance has an intersection with one of the interval_list
    Args:
        target (an Interval instance):
        interval_list (a list of Interval instances):

    Returns:
        Boolean

    """
    for interval in interval_list:
        if Interval.intersection(target, interval) is not None:
            return True
    return False


def interval_inclusion_intervallist(target, interval_list):
    """
    Check whether an Interval instance has an intersection with one of the interval_list
    Args:
        target (an Interval instance):
        interval_list (a list of Interval instances):

    Returns:
        Boolean

    """
    for interval in interval_list:
        if Interval.inclusion(target, interval):
            return True
    return False


def get_conditions(literals, d):
    """
    This function return the accepting conditions according to the direction ``d''.
    Args:
        literals ( list of Literal instances):
        d (the direction):

    Returns:
        list of accepting conditions .

    """
    F = list()
    for literal in literals:
        if d == "left":
            if isinstance(literal, BinaryLiteral):
                if literal.operator.right_value == Decimal("inf") and literal.operator.name == "Since":
                    F.append(literal)
            else:
                if isinstance(literal, Literal) and len(literal.operators) == 1 and (
                        literal.operators[0].right_value == Decimal("inf")) and literal.operators[0].name == "Boxminus":
                    F.append(literal)

        else:
            if isinstance(literal, BinaryLiteral):
                if literal.operator.left_value == Decimal("inf") and literal.operator.opName == "U":
                    F.append(literal)
            else:
                if isinstance(literal, Literal) and len(literal.operators) == 1 and (
                        literal.operators[0].right_value == Decimal("inf")) and literal.operators[0].name == "Boxplus":
                    F.append(literal)
    return F


def get_initial_ruler_intervals(points, left_border, right_border, gcd):
    """
    This function build initial intervals used for creating a AutomataWindow instance.
    Args:
        left_border (float): the most left minimum value  of the Window
        right_border(float):  the most right maximum value  of the Window
        gcd (float):
    Returns:
    """

    points = list(points)
    points.sort()

    left_end_point = points[0]
    right_end_point = points[0] - gcd

    pattern = set()
    pattern.add(left_end_point)
    pattern.add(right_end_point)
    # the pattern's range [x, x+gcd], all other points will fall into this range
    for item in points:
        a, b = divmod(item-left_end_point, gcd)
        if b == 0:
            continue
        else:
            pattern.add(b)

    pattern = list(pattern)
    pattern.sort()
    left_initial_intervals = []
    for i in range(len(pattern)):
        if i == len(pattern) - 1:
            left_initial_intervals.append(
                Interval(pattern[i], pattern[i], False, False))
        else:
            left_initial_intervals.append(
                Interval(pattern[i], pattern[i], False, False))
            left_initial_intervals.append(
                Interval(pattern[i], pattern[i + 1], True, True))

    left_end_point = points[-1]
    right_end_point = points[-1] + gcd

    pattern = set()
    pattern.add(left_end_point)
    pattern.add(right_end_point)
    # the pattern's range [x, x+gcd], all other points will fall into this range
    for item in points:
        a, b = divmod(item - left_end_point, gcd)
        if b == 0:
            continue
        else:
            pattern.add(b)

    pattern = list(pattern)
    pattern.sort()
    pattern_len = [pattern[i] - pattern[i - 1] for i in range(1, len(pattern))]
    pattern_num = len(pattern_len)

    right_initial_intervals = []
    for i in range(len(pattern)):
        if i == len(pattern) - 1:
            right_initial_intervals.append(
                Interval(pattern[i], pattern[i], False, False))
        else:
            right_initial_intervals.append(
                Interval(pattern[i], pattern[i], False, False))
            right_initial_intervals.append(
                Interval(pattern[i], pattern[i + 1], True, True))

    return left_initial_intervals, right_initial_intervals, pattern_num, pattern_len


def get_dataset_points_x(D, min_x_flag=False):
    """
    This function return all distinct points and the maximum point.
    Args:
        D (dictionary of dictionary):  contain all facts.

    Returns:
         (set, int)
    """
    max_x = decimal.Decimal("-inf")
    min_x = decimal.Decimal("+inf")
    points = set()
    for predicate in D:
        if type(D[predicate]) == list:
            for interval in D[predicate]:
                if interval.left_value not in [Decimal("inf"), Decimal("-inf")]:
                    points.add(interval.left_value)
                    max_x = max(max_x, interval.left_value)
                    min_x = min(min_x, interval.left_value)
                if interval.right_value not in [Decimal("inf"), Decimal("-inf")]:
                    points.add(interval.right_value)
                    max_x = max(max_x, interval.right_value)
                    min_x = min(min_x, interval.right_value)
        else:
            for entity in D[predicate]:
                for interval in D[predicate][entity]:
                    if interval.left_value not in [Decimal("inf"), Decimal("-inf")]:
                        points.add(interval.left_value)
                        max_x = max(max_x, interval.left_value)
                        min_x = min(min_x, interval.left_value)
                    if interval.right_value not in [Decimal("inf"), Decimal("-inf")]:
                        points.add(interval.right_value)
                        max_x = max(max_x, interval.right_value)
                        min_x = min(min_x, interval.left_value)

    points = list(points)
    points.sort()
    if min_x_flag:
        return points, min_x, max_x
    else:
        return points, max_x


def get_gcd(program):
    """
    This function calculate the maximum value in the program and  the great common divisor.
    Args:
        program (list of Rule instances):

    Returns:
       (float, float)
    """
    numbers = set()
    for rule in program:
        body = rule.body
        for literal in body + [rule.head]:
            if isinstance(literal, Atom):
                continue

            elif isinstance(literal, BinaryLiteral):
                if literal.operator.interval.left_value not in [Decimal("inf"), Decimal("-inf")]:
                    numbers.add(literal.operator.interval.left_value)
                if literal.operator.interval.right_value not in [Decimal("inf"), Decimal("-inf")]:
                    numbers.add(literal.operator.interval.right_value)

            else:
                for operator in literal.operators:
                    if operator.interval.left_value not in [Decimal("inf"), Decimal("-inf")]:
                        numbers.add(operator.interval.left_value)
                    if operator.interval.right_value not in [Decimal("inf"), Decimal("-inf")]:
                        numbers.add(operator.interval.right_value )

    precisions = [len(str(item).split(".")[1]) for item in numbers if len(str(item).split(".")) == 2]
    if len(precisions) == 0 and len(numbers) != 0:
        return max(numbers), reduce(lambda x, y: math.gcd(int(x), int(y)), numbers)
    elif len(precisions) == 0 and len(numbers) == 0:
        return 1, 1
    else:
        aug = 10 ** max(precisions)
        aug_numbers = [item*aug for item in numbers]
        gcd = reduce(lambda x, y: math.gcd(int(x), int(y)), aug_numbers) / aug
        gcd = decimal.Decimal(gcd)
        return max(numbers), gcd


def window_contain_accepting_conditions(F, W, subset_conditions, d):
    """
    This function judge whether the given AutomataWindow instance contains some accepting conditions
    Args:
        F (list of Literal instances):
        W (AutomtataWindow instance):
        subset_conditions:
        d:
    Returns:
    """
    ruler_intervals = W.ruler_intervals
    ruler_intervals_literals = W.ruler_intervals_literals

    for ruler_interval in ruler_intervals:
       for condition in F:
          if condition in ruler_intervals_literals[ruler_interval]:
             subset_conditions.add(condition)

          elif isinstance(condition, Literal):
             if condition.atom not in ruler_intervals_literals[ruler_interval]:
                if d == "left":
                   operator1 = Operator("Boxminus", Interval(0, Decimal("inf"), False, True))
                   literal1 = Literal(condition.atom, [operator1])
                   operator2 = Operator("Boxminus", Interval(0, Decimal("inf"), True, True))
                   literal2 = Literal(condition.atom, [operator2])
                   if literal1 in F:
                      subset_conditions.add(literal1)
                   if literal2 in F:
                      subset_conditions.add(literal2)
                else:
                    operator1 = Operator("Boxplus", Interval(0, Decimal("inf"), False, True))
                    literal1 = Literal(condition.atom, [operator1])
                    operator2 = Operator("Boxplus", Interval(0, Decimal("inf"), True, True))
                    literal2 = Literal(condition.atom, [operator2])
                    if literal1 in F:
                       subset_conditions.add(literal1)
                    if literal2 in F:
                       subset_conditions.add(literal2)

          elif isinstance(condition, BinaryLiteral):
               if condition.right.atom in ruler_intervals_literals[ruler_interval]:
                    subset_conditions.add(condition)

       if len(subset_conditions) != 0:
           return True
       else:
           return False


def extract_dataset(D, involved_predicates, automata_predicates=[]):
    """
    Extract Atom instances from the dataset and construct the corresponding unbounded literals.
    Args:
        D (dictionary of dictionary):
        automata_predicates (list of str): store predicates needed for Automata
        exclude (list of str): contain some predicat names that are not needed to construct unbounded literals.

    Returns:
        (list, list)
    """
    atoms = []
    unbounded_literals = []
    for predicate in involved_predicates:
        if predicate in D:
            if type(D[predicate]) == list:
                atom = Atom(predicate)
                atoms.append(atom)
                if predicate in automata_predicates:
                    literal = Literal(Atom(predicate),
                                       [Operator("Boxminus", Interval(0, Decimal("inf"), False, True))])
                    unbounded_literals.append(literal)
                    # literal = Literal(Atom(predicate),
                    #                    [Operator("Boxminus", Interval(0, Decimal("inf"), True, True))])
                    # unbounded_literals.append(literal)

                    literal = Literal(Atom(predicate),
                                      [Operator("Boxmplus", Interval(0, Decimal("inf"), False, True))])
                    unbounded_literals.append(literal)
                    # literal = Literal(Atom(predicate),
                    #                   [Operator("Boxmplus", Interval(0, Decimal("inf"), True, True))])
                    # unbounded_literals.append(literal)
            else:
                for entity in D[predicate]:
                    atom = Atom(predicate, entity)
                    atoms.append(atom)
                    if predicate in automata_predicates:
                        literal = Literal(Atom(predicate, entity),
                                          [Operator("Boxminus", Interval(0, Decimal("inf"), False, True))])
                        unbounded_literals.append(literal)
                        # literal = Literal(Atom(predicate, entity),
                        #                   [Operator("Boxminus", Interval(0, Decimal("inf"), True, True))])
                        # unbounded_literals.append(literal)

                        literal = Literal(Atom(predicate, entity),
                                          [Operator("Boxplus", Interval(0, Decimal("inf"), False, True))])
                        unbounded_literals.append(literal)
                        # literal = Literal(Atom(predicate, entity),
                        #                   [Operator("Boxplus", Interval(0, Decimal("inf"), True, True))])
                        # unbounded_literals.append(literal)

    return atoms, unbounded_literals


def extract_program(program, involved_predicates):
    atoms = set()
    literals = set()
    for rule in program:
        head = rule.head
        if head.get_predicate() in involved_predicates:
            continue
        for literal in rule.body():
            if isinstance(literal, BinaryLiteral):
                if literal.left_atom.get_predicate() in involved_predicates or literal.right_atom.get_predicate() in involved_predicates:
                    literals.add(literal)
            elif isinstance(literal, Literal):
                if literal.get_predicate() in involved_predicates:
                    literals.append(literal)
            else:
                if literal.get_predicate() in involved_predicates:
                    atoms.add(literal)
    return literals



import rdflib
from collections import defaultdict
import os
# import argparse
# parser = argparse.ArgumentParser()
# parser.add_argument("--owl_path",  type=str, help="input the dir_path where owl files locate.")
# parser.add_argument("--out_dir", type=str, help="input the path for the converted datalog triplets.")
# args = parser.parse_args()


def extract_triplets(owl_path, out_dir):
    entities = dict() # the key is the entity ids
    ID_ = "ID"
    ID_cnt = 0

    predicates = defaultdict(int)
    a1 = "http://swat.cse.lehigh.edu/onto/univ-bench.owl#"
    ignores = ["http://www.w3.org/2002/07/owl#imports", a1 + "emailAddress", a1 + "telephone"]

    if owl_path.endswith("/"):
        owl_path = owl_path[:-1]

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    fileout = open(os.path.join(out_dir, os.path.split(owl_path)[1]), "w")
    for filename in os.listdir(owl_path):
        g = rdflib.Graph()
        g.load(os.path.join(owl_path, filename))
        degree_intervals_dict = defaultdict(dict)
        for s, p, o in g:
            s, p, o = str(s), str(p), str(o)
            if p in ignores:
              continue

            if p == "http://www.w3.org/1999/02/22-rdf-syntax-ns#type":
                _prefix, _value = o.split("#")
                if "Ontology" in _value:
                    continue
                str_s = str(s).lower().strip()
                if str_s not in entities:
                    entities[str_s] = ID_ + str(ID_cnt)
                    str_s = ID_ + str(ID_cnt)
                    ID_cnt += 1
                else:
                    str_s = entities[str_s]

                fact = _value + "(" + str_s + ")"
                fileout.write(fact + "\n")
                predicates[_value] += 1

            else:
                _prefix, _value = p.split("#")
                str_s = str(s).lower()
                str_o = str(o).lower()
                if str_s not in entities:
                    entities[str_s] = ID_ + str(ID_cnt)
                    str_s = ID_ + str(ID_cnt)
                    ID_cnt += 1
                else:
                    str_s = entities[str_s]

                if str_o not in entities:
                    entities[str_o] = ID_ + str(ID_cnt)
                    str_o = ID_ + str(ID_cnt)
                    ID_cnt += 1
                else:
                    str_o = entities[str_o]

                fact = _value + "(" + str_s + "," + str_o + ")"
                fileout.write(fact + "\n")
                predicates[_value] += 1

    with open(os.path.join(out_dir, "statistics.txt"), "w") as file:
        number_triplets = 0
        for predicate, num in predicates.items():
            number_triplets += num
            file.write(predicate + ":" + str(num) + "\n")

        file.write("The number of unique entities:" + str(len(entities)) + "\n")
        file.write("The number of triplets:" + str(number_triplets))








from meteor_reasoner.utils.loader import *


class CycleFinder:
    def __init__(self, program):
        self.predicate2id = {}
        self.id2predicate = {}
        self.loop = list()
        self.graph = defaultdict(list)
        self.SCC_components = []
        self.Time = 0
        self.program = program
        self.build(program)
        self.cycle_generate()

    def add(self, predicate):
        if predicate not in self.predicate2id:
            id = len(self.predicate2id)
            self.predicate2id[predicate] = id
            self.id2predicate[id] = predicate

    def addedge(self, predicate1, predicate2):
        self.add(predicate1)
        self.add(predicate2)
        id1 = self.predicate2id[predicate1]
        id2 = self.predicate2id[predicate2]
        self.graph[id1].append(id2)

    def is_reachable(self, s, d):
        visited = [False] * self.V
        queue = []

        queue.append(s)
        visited[s] = True
        while queue:
            n = queue.pop(0)
            if n == d:
                return True
            for i in self.graph[n]:
                if visited[i] == False:
                    queue.append(i)
                    visited[i] = True

        return False

    def SCCUtil(self, u, low, disc, stackMember, st):
        disc[u] = self.Time
        low[u] = self.Time
        self.Time += 1
        stackMember[u] = True
        st.append(u)
        for v in self.graph[u]:
            if disc[v] == -1:
                self.SCCUtil(v, low, disc, stackMember, st)
                low[u] = min(low[u], low[v])

            elif stackMember[v] == True:
                low[u] = min(low[u], disc[v])

        w = -1
        if low[u] == disc[u]:
            tmp = []
            while w != u:
                w = st.pop()
                tmp.append(w)
                stackMember[w] = False

            self.SCC_components.append(tmp)

    def SCC(self):
        disc = [-1] * (self.V)
        low = [-1] * (self.V)
        stackMember = [False] * (self.V)
        st = []
        for i in range(self.V):
            if disc[i] == -1:
                self.SCCUtil(i, low, disc, stackMember, st)

    def build(self, program):
        for rule in program:
            head_predicate = rule.head.get_predicate()
            self.add(head_predicate)
            for literal in rule.body:
                if isinstance(literal, BinaryLiteral):
                    self.add(literal.left_literal.get_predicate())
                    self.add(literal.right_literal.get_predicate())
                else:
                   self.add(literal.get_predicate())

        for rule in program:
            head_predicate = rule.head.get_predicate()
            for literal in rule.body:
                if isinstance(literal, Literal) or isinstance(literal, Atom):
                    if literal.get_predicate() == head_predicate:
                        self.loop.append([self.predicate2id[head_predicate]])
                    self.addedge(literal.get_predicate(), head_predicate)
                else:
                    if head_predicate == literal.left_literal.get_predicate():
                        self.loop.append([self.predicate2id[head_predicate]])
                    if head_predicate == literal.right_literal.get_predicate():
                        self.loop.append([self.predicate2id[head_predicate]])

                    self.addedge(literal.left_literal.get_predicate(), head_predicate)
                    self.addedge(literal.right_literal.get_predicate(), head_predicate)

        self.V = len(self.predicate2id)


    def get_revevant_rules(self, predicate):
        relevant_rules = set()
        r_prime_body_prime = []
        old_len = len(relevant_rules)
        while True:
            for rule in self.program:
                if rule.head.get_predicate() in [predicate, "Bottom"] or rule in relevant_rules:
                    relevant_rules.add(rule)
                    for literal in rule.body:
                        if not isinstance(literal, BinaryLiteral):
                            r_prime_body_prime.append(literal.get_predicate())
                        else:
                            r_prime_body_prime.append(literal.left_literal.get_predicate())
                            r_prime_body_prime.append(literal.right_literal.get_predicate())

            for rule in self.program:
                head_predicate = rule.head.get_predicate()
                if head_predicate in r_prime_body_prime:
                    relevant_rules.add(rule)

            if len(set(relevant_rules)) > old_len:
                old_len = len(set(relevant_rules))
                continue

            else:
                break

        return list(relevant_rules)

    def cycle_generate(self):
        self.SCC()
        for scc in self.SCC_components:
            if len(scc) > 1:
                self.loop.append([id for id in scc])


    def get_non_recursive_predicates(self):
        non_recursive_predicates = set()
        for predicate in self.predicate2id:
            if not self.is_recursive_predicate(predicate):
                non_recursive_predicates.add(predicate)
        return [item for item in list(non_recursive_predicates) if item not in ["Top", "Bottom"]]


    def is_recursive_predicate(self, predicate):
        for cycle in self.loop:
            if predicate in self.predicate2id and self.predicate2id[predicate] in cycle:
                return True
            for middle_predicate in cycle:
                if self.is_reachable(middle_predicate, self.predicate2id[predicate]):
                    return True
        return False



from meteor_reasoner.graphutil.temporal_dependency_graph import *
from meteor_reasoner.graphutil.topological_sort import *


class Graph:
    def __init__(self, program):
        self.program = program
        self.head2rule = defaultdict(list)
        self.initialize()

    def initialize(self):
        self.build_head_map_rule()
        self.split_program()

    def split(self, rules, target_predicate):
        CF = CycleFinder(program=rules)
        CF.SCC()
        scc2id = defaultdict(int)
        id2scc = defaultdict(set)
        predicateid2sccid = dict()
        cnt = 0
        for scc in CF.SCC_components:
            scc2id[tuple(scc)] = cnt
            id2scc[cnt] = tuple(scc)
            for predicate in scc:
                predicateid2sccid[predicate] = cnt
            cnt += 1

        new_graph = defaultdict(list)
        graph = CF.graph

        for key, value in graph.items():
            key_id = predicateid2sccid[key]
            neighbors = []
            for neighbor in value:
                neighbor_id = predicateid2sccid[neighbor]
                if neighbor_id == key_id:
                    continue
                else:
                    neighbors.append(neighbor_id)
            if len(neighbors) > 0:
                new_graph[key_id] = new_graph[key_id] + neighbors

        TS = TopologicalSorting(new_graph, len(scc2id))
        sorted_scc = TS.topologicalSort()

        target_id = CF.predicate2id[target_predicate]
        non_rescursive_scc = set()
        automata_scc = set()

        target_predicate_scc_id_index = -1
        for i, scc_id in enumerate(sorted_scc):
            if target_id in id2scc[scc_id]:
                target_predicate_scc_id_index = i
                break
        if target_predicate_scc_id_index == -1:
            raise ValueError("No such predicate")

        first_cyclic_scc_id_index = -1
        for i, scc_id in enumerate(sorted_scc[:target_predicate_scc_id_index]):
              if len(id2scc[scc_id]) > 1:
                  first_cyclic_scc_id_index = i

        if first_cyclic_scc_id_index == -1:
            # no cycle before target_predicate_scc_id_index
            for scc_id in sorted_scc[:target_predicate_scc_id_index]:
                non_rescursive_scc = non_rescursive_scc.union(id2scc[scc_id])

            automata_scc = id2scc[sorted_scc[target_predicate_scc_id_index]]
        else:
            for scc_id in sorted_scc[:first_cyclic_scc_id_index]:
                non_rescursive_scc = non_rescursive_scc.union(id2scc[scc_id])
            for scc_id in sorted_scc[first_cyclic_scc_id_index:target_predicate_scc_id_index]:
                automata_scc = automata_scc.union(id2scc[scc_id])

        non_recursive_rules = set()
        automata_rules = set()

        for predicate_id in non_rescursive_scc:
            for rule in self.head2rule[CF.id2predicate[predicate_id]]:
                non_recursive_rules.add(rule)

        for predicate_id in automata_scc:
            for rule in self.head2rule[CF.id2predicate[predicate_id]]:
               flag = True
               for literal in rule.body:
                   if isinstance(literal, BinaryLiteral):
                       if literal.left_atom.get_predicate() not in ["Top", "Bottom"] and CF.predicate2id[literal.left_atom.get_predicate()] not in automata_scc:
                           flag = False
                       if literal.right_atom.get_predicate() not in ["Top", "Bottom"] and CF.predicate2id[literal.right_atom.get_predicate()] not in automata_scc:
                           flag = False
                   else:
                       if literal.get_predicate() not in automata_scc:
                           flag = False
               if flag:
                    automata_rules.add(rule)
               else:
                    non_recursive_rules.add(rule)

        return non_recursive_rules, automata_rules

    def build_head_map_rule(self):
        for rule in self.program:
            head = rule.head
            self.head2rule[head.get_predicate()].append(rule)

    def split_program(self):
        for rule in self.program:
            head_predicate = rule.head.get_predicate()
            self.cluster.add(head_predicate)
        for rule in self.program:
            head_predicate = rule.head.get_predicate()
            for literal in rule.body:
                 if (isinstance(literal, Literal) or isinstance(literal, Atom)) and literal.get_predicate() in self.cluster.predicate2id:
                    self.cluster.addedge(head_predicate, literal.get_predicate())
                 elif isinstance(literal, BinaryLiteral):
                     if literal.left_atom.get_predicate() in self.cluster.predicate2id:
                         if literal.left_atom.get_predicate() == head_predicate:
                             self.cluster.addedge(head_predicate, literal.left_atom.get_predicate() + "_")
                             self.cluster.addedge(literal.left_atom.get_predicate() + "_", head_predicate)
                         else:
                             self.cluster.addedge(literal.left_atom.get_predicate(), head_predicate)

                     if literal.right_atom.get_predicate() in self.cluster.predicate2id:
                         if literal.right_atom.get_predicate() == head_predicate:
                             self.cluster.addedge(head_predicate, literal.right_atom.get_predicate() + "_")
                             self.cluster.addedge(literal.right_atom.get_predicate() + "_", head_predicate)
                         else:
                             self.cluster.addedge(literal.right_atom.get_predicate(), head_predicate)

    def split_rules(self, target_predicate):
        predicates = self.cluster.get_connected_predicates(target_predicate)
        rules = set()
        for predicate in predicates:
            rules = rules.union(set(self.head2rule[predicate]))

        return self.split(rules, target_predicate)





from meteor_reasoner.classes.rule import *
from meteor_reasoner.materialization.ground import *
from meteor_reasoner.classes.term import Term
from collections import defaultdict
from meteor_reasoner.materialization.index_build import build_index
from meteor_reasoner.utils.operate_dataset import print_dataset
from meteor_reasoner.materialization.coalesce import coalescing_d


def naive_join(rule, D, delta_new, D_index=None):
    """
    This function implement the join operator when variables exist in the body of the rule.
    Args:
        rule (a Rule instance):
        delta_new (a dictionary object): store new materialized results.
        D (a dictionary object): a global variable storing all facts.
        common_fragment: Canonical Representation related

    Returns:
        None
    """
    head_entity = rule.head.get_entity()
    head_predicate = rule.head.get_predicate()
    literals = rule.body + rule.negative_body

    def ground_body(global_literal_index, delta, context):
        if global_literal_index == len(literals):
            T = []
            for i in range(len(rule.body)):
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    grounded_literal.set_entity(delta[i])
                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])
                t = apply(grounded_literal, D)
                if len(t) == 0:
                    break
                else:
                    T.append(t)
            n_T = []
            for i in range(len(rule.body), len(literals)):
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    grounded_literal.set_entity(delta[i])
                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])

                t = apply(grounded_literal, D)
                if len(t) == 0:
                    break
                else:
                    n_T.append(t)
            if len(n_T) > 0 and len(n_T) == len(rule.negative_body_atoms):
                n_T = interval_merge(T)
            else:
                n_T = []
            if len(head_entity) > 1 or head_entity[0].name != "nan":
                replaced_head_entity = []
                for term in head_entity:
                    if term.type == "constant":
                        replaced_head_entity.append(term)
                    else:
                        if term.name not in context:
                            raise ValueError("Head variables in Rule({}) do not appear in the body".format(str(rule)))
                        else:
                            new_term = Term.new_term(context[term.name])
                            replaced_head_entity.append(new_term)

                replaced_head_entity = tuple(replaced_head_entity)
            else:
                replaced_head_entity = head_entity

            if len(T) == len(literals):
                T = interval_merge(T)
                exclude_t = []
                if len(T) != 0 and len(n_T) != 0:
                    exclude_t = interval_merge([T, n_T])
                if len(exclude_t) != 0:
                    T = Interval.diff(T, exclude_t)
                if len(T) != 0:
                    if not isinstance(rule.head, Atom):
                        tmp_D = defaultdict(lambda: defaultdict(list))
                        tmp_D[head_predicate][replaced_head_entity] = T
                        tmp_head = copy.deepcopy(rule.head)
                        tmp_head.set_entity(replaced_head_entity)
                        T = reverse_apply(tmp_head, tmp_D)
                    delta_new[head_predicate][replaced_head_entity] += T
        else:
            current_literal = copy.deepcopy(literals[global_literal_index])
            if not isinstance(current_literal, BinaryLiteral):
                if current_literal.get_predicate() in ["Bottom", "Top"]:
                    ground_body(global_literal_index+1, delta, context)
                else:
                    for tmp_entity, tmp_context in ground_generator(current_literal, context, D, D_index):
                        tmp_delata = {global_literal_index: [tmp_entity]}
                        ground_body(global_literal_index+1, {**delta, **tmp_delata}, {**context, **tmp_context})
            else:
                left_predicate = current_literal.left_literal.get_predicate()
                right_predicate = current_literal.right_literal.get_predicate()

                if left_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_context in ground_generator(current_literal.right_literal, context,  D, D_index):
                        tmp_delta = {global_literal_index: [tmp_entity]}
                        ground_body(global_literal_index + 1, {**delta, **tmp_delta}, {**context, **tmp_context})

                elif right_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_context in ground_generator(current_literal.left_literal, context, D, D_index):
                        tmp_delta = {global_literal_index: [tmp_entity]}
                        ground_body(global_literal_index + 1, {**delta, **tmp_delta}, {**context, **tmp_context})

                else:
                    for left_entity, tmp_context1 in ground_generator(current_literal.left_literal, context, D):
                        for right_entity, tmp_context2 in ground_generator(current_literal.right_literal.atom,{**context, **tmp_context1}, D):
                            tmp_delta = {global_literal_index: [left_entity, right_entity]}
                            ground_body(global_literal_index + 1, {**delta, **tmp_delta}, {**context, **tmp_context1, **tmp_context2})

    ground_body(0, {}, dict())


if __name__ == "__main__":
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike"), Term("nick")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D["P"][tuple([Term("a")])] = [Interval(1, 1, False, False)]
    D_index = build_index(D)

    Delta = defaultdict(lambda: defaultdict(list))
    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("Y", "variable"), Term("X", "variable")])),
                        [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])

    body = [literal_a, literal_b]

    new_literal = BinaryLiteral(Atom("A", tuple([Term("X", "variable"), Term("Y", "variable")])),
                                Atom("B", tuple([Term("nan")])), Operator("Since", Interval(1, 2, False, False)))
    body.append(new_literal)
    literal_p = Literal(Atom("P", tuple([Term("X", "variable")])),
                        [Operator("Diamondminus", Interval(1, 1, False, False))])
    head_p = Atom("P", tuple([Term("X", "variable")]))

    rule = Rule(head_p, [literal_p])
    i = 0
    delta_old = D
    while i < 5:
        i += 1
        delta_new = defaultdict(lambda: defaultdict(list))
        naive_join(rule, D=D, delta_new=delta_new, D_index=D_index)
        print("new:")
        print_dataset(delta_new)
        for predicate in delta_new:
            if predicate not in D:
                D[predicate] = delta_new[predicate]
            else:
                for entity in delta_new[predicate]:
                    D[predicate][entity] = D[predicate][entity] + delta_new[predicate][entity]
        coalescing_d(D)
        delta_old = delta_new














import subprocess
import os
import shutil
# import argparse
#
# parser = argparse.ArgumentParser()
# parser.add_argument("--univ_num", type=int, help= "Input the number of universities you want to generate.")
# parser.add_argument("--dir_name", type=str, help= "Input the directory path used for the generated owl files.")
# args = parser.parse_args()


def generate(univ_num=1, dir_name="./owl_data"):
    onto="http://swat.cse.lehigh.edu/onto/univ-bench.owl"
    cmd = "java edu.lehigh.swat.bench.uba.Generator -univ {} -onto {}"

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    os.chdir(dir_name)
    subprocess.call(cmd.format(univ_num, onto), shell=True)
    os.chdir("../")
    subprocess.call("rm *log.txt*", shell=True)

    for filename in os.listdir("./"):
        if "University" in filename and ".owl" in filename:
            start_index = filename.index("University")
            new_name = filename[start_index:]
            foldername = filename[0:start_index-1]
            if not os.path.exists('./' + foldername):
                os.makedirs('./' + foldername)
            shutil.move("./" + filename, os.path.join(foldername, new_name))


if __name__ == "__main__":
    generate()


from meteor_reasoner.materialization.seminaive_join import *
from meteor_reasoner.materialization.naive_join import *
from collections import defaultdict


def naive_immediate_consequence_operator(rules, D, D_index, must_literals=None):
    delta_new = defaultdict(lambda: defaultdict(list))
    for i, rule in enumerate(rules):
        naive_join(rule, D, delta_new, D_index, must_literals=must_literals)
    return delta_new


def seminaive_immediate_consequence_operator(rules, D, D_index, delta_old=None, must_literals=None):
    delta_new = defaultdict(lambda: defaultdict(list))
    for i, rule in enumerate(rules):
        seminaive_join(rule, D, delta_old, delta_new, D_index=D_index, must_literals=must_literals)

    return delta_new



from meteor_reasoner.graphutil.multigraph import TemporalDependencyGraph


def test_multigraph():
    G = TemporalDependencyGraph()
    G.add_edge("Vaccinated", "Immune", [21,28])
    G.add_edge("NoSympt", "Immune", [0, 10])
    G.add_edge("Infected", "Immune", [10, 183])
    G.add_edge("NegTest", "Immune", [0, 0])
    G.add_edge("Immune", "NegTest", [0, 5])

    print(G.is_cyclic())

    G = TemporalDependencyGraph()
    G.add_edge("NotInf", "Susc", [0,183])
    G.add_edge("NotVacc", "Susc", [0, 365])
    G.add_edge("FirstSympt", "GetsInf", [5, 5])
    G.add_edge("GetsInf", "FirstSympt", [-5, -5])
    G.add_edge("Over65", "FirstSympt", [0, 0])
    G.add_edge("NoMask", "GetsInf", [1/12, 1/12])
    print(G.is_cyclic())



test_multigraph()


from meteor_reasoner.utils.loader import *
from meteor_reasoner.canonical.find_common_fragment import CanonicalRepresentation, find_left_right_periods, fact_entailment
from meteor_reasoner.canonical.calculate_w import get_w


def is_entail(dataset, program, fact):
    """
    Args:
        dataset: a list of string of the form A(a,b)@[1,2]
        program: a list of string of the form A:- B, C
        fact: a string of the form A(a,b)@[1,2]
    Returns:
        Boolean value
    """
    D = load_dataset(dataset)
    program = load_program(program)
    predicate, entity, interval = parse_str_fact(fact)
    fact = Atom(predicate, entity, interval)

    # calculate w
    # will change the program, so we need to make a copy
    t_program = copy.deepcopy(program)
    w = 2 * get_w(t_program)
    # calculate the minimum value and maximum value in the dataset
    maximum_number = -10
    minimum_number = 100
    for predicate in D:
        for entity in D[predicate]:
            for interval in D[predicate][entity]:
                maximum_number = max(interval.right_value, maximum_number)
                minimum_number = min(interval.left_value, minimum_number)

    # calculate the canonical representation
    CR = CanonicalRepresentation(D, program)
    D1, common, varrho_left, left_period, left_len, varrho_right, right_period, right_len = find_left_right_periods(CR, w, fact)

    flag = fact_entailment(D1, fact, common, left_period, left_len, right_period, right_len)
    return flag


if __name__ == "__main__":
    dataset = [
        "HighTemperature(sensor1) @ [0, 27.5]",
        "HighTemperature(sensor2) @ 3.5",
        "HighTemperature(sensor2) @ 5.1",
        "HighTemperature(sensor2) @ 10",
        "HighTemperature(sensor2) @ 14.7",
        "HighTemperature(sensor2) @ 20"
    ]
    # dataset = ["A(a)@[1,2]", "B(a)@[1,2]"]
    program = [
        "Overheat(X):- ALWAYS[-10,0]HighTemperature(X)",
        "Overheat(X): - ALWAYS[-20,0]SOMETIME[-5.5,0]HighTemperature(X)",
        "Alert: - Overheat(X), Overheat(Y)"
    ]
    # program = ["C(X):-A(X),B(X)"]
    fact = "Alert@25"
    print(is_entail(dataset, program, fact))



from meteor_reasoner.classes.rule import *
from meteor_reasoner.classes.term import *
from meteor_reasoner.classes.interval import *
from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.atom import *


from meteor_reasoner.utils.ruler_interval import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.classes.literal import Literal, BinaryLiteral
import decimal, copy
from meteor_reasoner.canonical.calculate_w import get_w


class CanonicalRepresentation:
    def __init__(self, D, Program):
        self.D = D
        self.D_index = defaultdict(lambda : defaultdict(list))
        self.Program = Program

    def is_bounded_program(self):
        for rule in self.Program:
            for literal in rule.body:
                if isinstance(literal, Literal):
                    for op in literal.operators:
                        if op.right_value in [decimal.Decimal("+inf")]:
                            return False
                elif isinstance(literal, BinaryLiteral):
                    if literal.operator.right_value in [decimal.Decimal("+inf")]:
                        return False
        return True

    def initilization(self):
        t_program = copy.deepcopy(self.Program)
        self.w = 2 * get_w(t_program)
        coalescing_d(self.D)
        build_index(self.D,  self.D_index)
        self.points, self.min_x, self.max_x = get_dataset_points_x(self.D, min_x_flag=True)

        print("The w is:{}".format(self.w))
        print("The maximum number in the database:{}".format(self.max_x))
        print("The minimum number in the database:{}".format(self.min_x))
        self.base_interval = Interval(self.min_x, self.max_x, False, False)
        self.z, self.gcd = get_gcd(self.Program)
        self.left_initial_ruler_intervals, self.right_initial_ruler_intervals, self.pattern_num, self.pattern_len = get_initial_ruler_intervals(self.points[:], left_border= self.min_x-self.w, right_border=self.max_x+self.w, gcd=self.gcd)

        self.left_dict, self.right_dict = construct_left_right_pattern(self.points, self.gcd)



import os
import random

import argparse

# parser = argparse.ArgumentParser(description='Process some integers.')
# parser.add_argument('integers', metavar='N', type=int, choices=[1,2,3], nargs='*',
#                     help='an integer for the accumulator')
# parser.add_argument('--sum', '-s', dest='accumulate', action='store_const',
#                     const=sum, default=max,
#                     help='sum the integers (default: find the max)')
#
# args = parser.parse_args()
# print(args.accumulate)
# print(args.accumulate(args.integers))
# exit()

parser = argparse.ArgumentParser()
parser.add_argument("--datalog_file_path", type=str, required=True)
parser.add_argument("--factnum", type=int, default=1000, required=True)
parser.add_argument("--intervalnum", type=int, default=1)
parser.add_argument("--unit", default=1, type=float)
parser.add_argument("--punctual", action="store_true", help="specify whether we only add punctual intervals")
parser.add_argument("--openbracket", action="store_true", help="specify whether we need to add open brackets")
parser.add_argument("--min_val", type=float, default=0)
parser.add_argument("--max_val", type=float, default=50)
args = parser.parse_args()


POINTS = []
cur_point = args.min_val
while cur_point <= args.max_val:
    POINTS.append(round(cur_point, 1))
    cur_point += args.unit

interval_num=args.intervalnum
current_num = 0
fileout = open(os.path.join(os.path.dirname(args.datalog_file_path), str(args.factnum)+".txt"), "w")

with open(args.datalog_file_path) as file:
    for line in file:
        if current_num == args.factnum:
            break

        elif args.factnum - current_num >= interval_num:
            current_num += interval_num
            for _ in range(interval_num):
                if args.punctual:
                    val = random.choice(POINTS)
                    left, right = val, val
                else:
                    left, right = random.sample(POINTS, 2)
                    if left > right:
                        left, right = right, left
                if args.openbracket:
                    left_bracket = random.choice(["(", "["])
                    right_bracket = random.choice([")", "]"])
                    new_fact = line.strip() + "@" + "{}{},{}{}".format(left_bracket, left, right, right_bracket)
                else:
                    new_fact = line.strip() + "@" + "[{},{}]".format(left, right)
                fileout.write(new_fact + "\n")
        else:
            current_num += args.factnum - current_num
            for _ in range(args.factnum - current_num):
                if args.punctual:
                    val = random.choice(POINTS)
                    left, right = val, val
                else:
                    left, right = random.sample(POINTS, 2)
                    if left > right:
                        left, right = right, left
                if args.openbracket:
                    left_bracket = random.choice(["(", "["])
                    right_bracket = random.choice([")", "]"])
                    new_fact = line.strip() + "@" + "{}{},{}{}".format(left_bracket, left, right, right_bracket)
                else:
                    new_fact = line.strip() + "@" + "[{},{}]".format(left, right)
                fileout.write(new_fact + "\n")


from meteor_reasoner.graphutil.graph_strengthening import *
from meteor_reasoner.classes import *
import copy


def test_strengthening():
    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False)),Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    literal_c = BinaryLiteral(copy.deepcopy(literal_a), copy.deepcopy(literal_b), Operator("Since", Interval(4, 5, False, False)))
    body = [literal_a, literal_b, literal_c]
    rule = Rule(head, body)
    print(rule)
    new_rule = transformation([rule])
    print(new_rule[0])


test_strengthening()


from meteor_reasoner.classes.interval import Interval


def test_interval():
    # test the add, sub, cicle_add, cicle_sub
    interval1 = Interval(7, 10, False, False)
    interval2 = Interval(2, 4, False, False)

    truth_interval = Interval(9, 14, False, False)
    # usually used from diamondminus
    interval = Interval.add(interval1, interval2)
    assert truth_interval == interval

    truth_interval = Interval(11, 12, False, False)
    # usually used from boxminus
    interval = Interval.circle_add(interval1, interval2)
    assert truth_interval == interval

    truth_interval = Interval(5, 6, False, False)
    # usually used from boxplus
    interval = Interval.circle_sub(interval1, interval2)
    assert truth_interval == interval

    truth_interval = Interval(3, 8, False, False)
    # usually used from diamondplus
    interval = Interval.sub(interval1, interval2)
    assert truth_interval == interval


def test_static():
    interval1 = Interval(3, 4,False, False)
    interval2 = Interval(3, 4, False, False)
    print(Interval.inclusion(interval1, interval2))

    interval1 = Interval(3, 4, False, False)
    interval2 = Interval(3.5, 3.8, False, False)
    results = Interval.diff(interval1, [interval2])
    print([str(res) for res in results])


test_static()


from meteor_reasoner.graphutil.temporal_dependency_graph import CycleFinder
from meteor_reasoner.utils.loader import load_program
from meteor_reasoner.classes.literal import BinaryLiteral

def test_cycle():
    rulepath = "/Users/dimmy/Desktop/MeTeoR/experiments/ISWC/programs/test.txt"
    with open(rulepath) as file:
        raw_program = file.readlines()
        program = load_program(raw_program)
        predicates = set()
        for rule in program:
            predicates.add(rule.head.get_predicate())
            for literal in rule.body:
                if isinstance(literal, BinaryLiteral):
                    predicates.add(literal.left_literal.get_predicate())
                    predicates.add(literal.right_literal.get_predicate())
                else:
                    predicates.add(literal.get_predicate())

    print("Number of predicates:", len(predicates))
    CF = CycleFinder(program=program)
    non_predicates = CF.get_non_recursive_predicates()

    print("Number of non-predicates:", len(non_predicates))
    for predicate in predicates:
        if predicate not in non_predicates:
              print(predicate)



from meteor_reasoner.utils.operate_dataset import *
from meteor_reasoner.materialization.materialize import *
import time

def test_naive_mat():
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["C"][tuple([Term("mike"), Term("jack")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D_index = defaultdict(lambda: defaultdict(list))
    build_index(D, D_index)

    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    body = [literal_a, literal_b]
    rule = Rule(head, body)
    Program = [rule]

    start_time = time.time()
    flag,_ = materialize(D, Program, K=10)
    end_time = time.time()
    print("flag:", flag)
    print("naive run time:", end_time-start_time)
    print("Naive:")
    print_dataset(D)


def test_seminaive_mat():
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["C"][tuple([Term("mike"), Term("jack")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D_index = defaultdict(lambda: defaultdict(list))
    build_index(D, D_index)

    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])
    body = [literal_a, literal_b]
    rule = Rule(head, body)
    Program = [rule]

    start_time = time.time()
    flag, _ = materialize(D, Program, seminaive=True, K=10)
    end_time = time.time()
    print("flag:", flag)
    print("seminaive run time:", end_time - start_time)
    print("Naive:")
    print_dataset(D)



import argparse
import os
parser = argparse.ArgumentParser()
parser.add_argument("--fixatom_datalogmtl_file_path", type=str)
parser.add_argument("--repeated_num", type=int, default=10)
parser.add_argument("--scale", default=50, type=float)
args = parser.parse_args()


lines = []
with open(args.fixatom_datalogmtl_file_path) as file:
   for line in file:
        a, b = line.split("@")
        b = b.strip()
        left_bracket = b[0]
        right_bracket = b[-1]
        left_point, right_point = b[1:-1].split(",")
        fact = "{}@{}{},{}{}"
        for i in range(args.repeated_num):
            new_fact = fact.format(a, left_bracket, float(left_point) + i*args.scale, float(right_point) + i*args.scale, right_bracket)
            lines.append(new_fact+"\n")

fix_num = int(len(lines) / args.repeated_num)
increased_num = int(len(lines))
with open(os.path.join(os.path.dirname(args.fixatom_datalogmtl_file_path), "fixatom_{}_{}".format(fix_num, increased_num)), "w") as file:
    for line in lines:
        file.write(line)






from meteor_reasoner.utils.parser import *
from collections import defaultdict


def load_dataset(lines):
    """
    Read string-like facts into a dictionary object.

    Args:
        lines (list of strings): a list of facts in the form of  A(x,y,z)@[1,2] or A@[1,2)

    Returns:
        A defaultdict object, in which the key is the predicate and the value is a dictionary (key is
        the entity and the value is a list of Interval instances) or a list of Interval instance when
        there is no entity.

    """
    D = defaultdict(lambda: defaultdict(list))
    for line in lines:
        line = line.strip().replace(" ","")
        if line == "":
            continue
        try:
          predicate, entity, interval = parse_str_fact(line)
        except:
            continue
        if predicate not in D:
            if entity:
               D[predicate][entity] = [interval]
            else:
               D[predicate] = [interval]
        else:
            if isinstance(D[predicate], list) and entity is not None:
                raise ValueError("One predicate can not have both entity and Null cases!")

            if not isinstance(D[predicate], list) and entity is None:
                raise ValueError("One predicate can not have both entity and Null cases!")

            if entity:
                if entity in D[predicate]:
                    D[predicate][entity].append(interval)
                else:
                    D[predicate][entity] = [interval]
            else:
                D[predicate].append(interval)


    return D


def load_program(rules):
    """
    Format each string-like rule into a rule instance.

    Args:
        rules (list of strings): each string represents a rule, e.g. A(X):- Boxminus[1,2]B(X)

    Returns:
        list of rule instances
    """
    program = []
    for line in rules:
        rule = parse_rule(line)
        program.append(rule)

    return program
